#pragma once

#define HOMOWEIGHT_BIN_PATH			"homoweight.bin"
#define WORDWEIGHT_BIN_PATH			"wordweight.bin"
#define PREDICT_BIN_PATH			"npredict.bin"
#define MORPH_MAIN_FILES			"morph.bin"
#define OPTIONS_FILE				"morph.options"

const uint32_t ErrorParadigmId			= 0xffffffff;
